/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 1999-2013 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */

#define  __NO_VERSION__
#include "nv-misc.h"

#include "os-interface.h"
#include "nv-linux.h"

RM_STATUS NV_API_CALL nv_dma_map_pages(
    nv_state_t *nv,
    NvU64       page_count,
    NvU64      *pte_array,
    void      **priv
)
{
    RM_STATUS status;
    NvU64 i, j;
    nv_linux_state_t *nvl = NV_GET_NVL_FROM_NV_STATE(nv);
    nv_dma_map_t *dma_map = NULL;

    if (priv == NULL)
    {
        /*
         * IOMMU path has not been implemented yet to handle
         * anything except a nv_dma_map_t as the priv argument.
         */
        return RM_ERR_NOT_SUPPORTED;
    }

    if (page_count > os_get_num_phys_pages())
    {
        nv_printf(NV_DBG_ERRORS,
                "NVRM: DMA mapping request too large!\n");
        return RM_ERR_INVALID_REQUEST;
    }

    status = os_alloc_mem((void **)&dma_map, sizeof(nv_dma_map_t));
    if (status != RM_OK)
    {
        nv_printf(NV_DBG_ERRORS,
                "NVRM: Failed to allocate nv_dma_map_t!\n");
        return status;
    }

    dma_map->user_pages = *priv;
    dma_map->dev = nvl->dev;

    for (i = 0; i < page_count; i++)
    {
        pte_array[i] = pci_map_page(dma_map->dev, dma_map->user_pages[i], 0,
                PAGE_SIZE,
                PCI_DMA_BIDIRECTIONAL);
        if (NV_PCI_DMA_MAPPING_ERROR(dma_map->dev, pte_array[i]) ||
            (!IS_DMA_ADDRESSABLE(nv, pte_array[i])))
        {
            nv_printf(NV_DBG_ERRORS,
                    "NVRM: failed to create a DMA mapping!\n");
            if (!IS_DMA_ADDRESSABLE(nv, pte_array[i]))
            {
                nv_printf(NV_DBG_ERRORS,
                        "NVRM: DMA address not in addressable range of device "
                        "%04x:%02x:%02x (0x%llx, 0x%llx-0x%llx)\n",
                        NV_PCI_DOMAIN_NUMBER(dma_map->dev),
                        NV_PCI_BUS_NUMBER(dma_map->dev),
                        NV_PCI_SLOT_NUMBER(dma_map->dev),
                        pte_array[i], nv->dma_addressable_start,
                        nv->dma_addressable_limit);
                status = RM_ERR_INVALID_ADDRESS;
            }
            else
            {
                status = RM_ERR_OPERATING_SYSTEM;
            }

            for (j = 0; j < i; j++)
            {
                pci_unmap_page(dma_map->dev, pte_array[j],
                        PAGE_SIZE, PCI_DMA_BIDIRECTIONAL);
            }

            os_free_mem(dma_map);
            return status;
        }
    }

    *priv = dma_map;

    return RM_OK;
}

RM_STATUS NV_API_CALL nv_dma_unmap_pages(
    nv_state_t *nv,
    NvU64       page_count,
    NvU64      *pte_array,
    void      **priv
)
{
    NvU64 i;
    nv_dma_map_t *dma_map;

    if (priv == NULL)
    {
        /*
         * IOMMU path has not been implemented yet to handle
         * anything except a nv_dma_map_t as the priv argument.
         */
        return RM_ERR_NOT_SUPPORTED;
    }

    dma_map = *priv;

    if (page_count > os_get_num_phys_pages())
    {
        nv_printf(NV_DBG_ERRORS,
                "NVRM: DMA unmapping request too large!\n");
        return RM_ERR_INVALID_REQUEST;
    }

    for (i = 0; i < page_count; i++)
    {
        pci_unmap_page(dma_map->dev, pte_array[i],
                PAGE_SIZE, PCI_DMA_BIDIRECTIONAL);
    }

    *priv = dma_map->user_pages;
    os_free_mem(dma_map);

    return RM_OK;
}
